from pandas import DataFrame, Series, to_datetime, to_timedelta


def extract_current_day_date(series_dt: Series) -> Series:
    # the result should be returned with 0-ed time information
    return series_dt.dt.normalize()


def extract_first_day_of_iso_week(series_dt: Series) -> Series:
    dayofweek = series_dt.dt.isocalendar().day
    # we subtract a number of days corresponding to(dayOfWeek - 1)
    result = series_dt - to_timedelta(dayofweek - 1, unit="d")
    # the result should be returned with 0-ed time information
    return to_datetime(result.dt.date)


def extract_first_day_of_week(series_dt: Series) -> Series:
    # dayofweek should be between 1 (sunday) and 7 (saturday)
    dayofweek = (series_dt.dt.dayofweek + 2) % 7
    dayofweek = dayofweek.replace({0: 7})
    # we subtract a number of days corresponding to(dayOfWeek - 1)
    result = series_dt - to_timedelta(dayofweek - 1, unit="d")
    # the result should be returned with 0-ed time information
    return to_datetime(result.dt.date)


def extract_first_day_of_month(series_dt: Series) -> Series:
    series_dt_accessor = series_dt.dt
    return to_datetime(DataFrame({"year": series_dt_accessor.year, "month": series_dt_accessor.month, "day": 1}))


def extract_first_day_of_quarter(series_dt: Series) -> Series:
    series_dt_accessor = series_dt.dt
    return to_datetime(
        DataFrame(
            {
                "year": series_dt_accessor.year,
                "month": 3 * ((series_dt_accessor.month - 1) // 3) + 1,
                "day": 1,
            }
        )
    )


def extract_first_day_of_year(series_dt: Series) -> Series:
    return to_datetime(DataFrame({"year": series_dt.dt.year, "month": 1, "day": 1}))



import ast
import tokenize
from collections.abc import Generator, Iterator
from io import BytesIO

from . import types


class ASTError(Exception):
    """Base exception for all AST errors"""


class UnclosedColumnName(ASTError):
    """Column name was not closed"""


class EmptyColumnName(ASTError):
    """Column name was not closed"""


class UnexpectedToken(ASTError):
    """Unexpected token"""


class InvalidFormula(ASTError):
    """Invalid formula"""


class UnsupportedOperator(ASTError):
    """Unsupported operator"""


class UnsupportedExpression(ASTError):
    """Unsupported expression"""


class UnsupportedConstant(ASTError):
    """Unsupported constant"""


class FormulaParser:
    def __init__(self, formula: str) -> None:
        self._formula = formula
        self._columns: dict[str, str] = {}

    def _iterate_tokens(self, tokens: Iterator[tokenize.TokenInfo]) -> Generator[str, None, None]:
        """This function iterates over tokens, sanitizes and yields them."""

        def next_token():
            try:
                return next(tokens)
            except StopIteration:
                return None

        def parse_col_name(prev_token: tokenize.TokenInfo) -> str:
            # NOTE: position is expressed as a (posLine, posColumn) tuple. We assume that the
            # formula is written on a single line
            start = prev_token.end[1]
            # Skipping the current token, which is [
            while tok := next_token():
                # We reached the end of the column name. Store an alias for it and yield the alias
                if tok.type == tokenize.OP and tok.string == "]":
                    end = tok.start[1]
                    if end - start < 1:
                        raise EmptyColumnName(f"Got an empty column name at {prev_token.start[1]}")
                    col_name = f"__VQB_COL__{len(self._columns)}"
                    self._columns[col_name] = self._formula[start:end]
                    return col_name
                prev_token = tok
            raise UnclosedColumnName(f"Expected column to be closed near {prev_token.string} at {prev_token.start[1]}")

        while token := next_token():
            if token.type in (
                tokenize.ENCODING,
                tokenize.INDENT,
                tokenize.NEWLINE,
                tokenize.ENDMARKER,
            ):
                # Those are whitespace tokens we don't care about, so not yielding them
                continue
            elif token.type == tokenize.OP and token.string == "[":
                # Square brackets are delimiters for column names, so here we read everything until
                # the closing bracket
                yield parse_col_name(token)
            else:
                yield token.string

    def sanitize_formula(self) -> str:
        """Removes [] around column names and aliases them.

        This is needed because stuff parsed with the STL's ast module needs to be valid python code
        """
        self._columns = {}
        # Stripping because strings starting with whitespace raise UnexpectedIndent when parsed by
        # the ast module
        return " ".join(self._iterate_tokens(tokenize.tokenize(BytesIO(self._formula.encode()).readline))).strip()

    @staticmethod
    def _operator_from_ast_op(op: ast.operator) -> types.Operator:
        """Returns one of our known operators from an operator of the ast module.

        Raise an exception in case of an unsupported operator.
        """
        match op:
            case ast.Add():
                return types.Operator.ADD
            case ast.Sub():
                return types.Operator.SUB
            case ast.Mult():
                return types.Operator.MUL
            case ast.Div():
                return types.Operator.DIV
            case ast.Mod():
                return types.Operator.MOD
            case _:
                raise UnsupportedOperator(f"Unsupported operator: {op}")

    def _build_name(self, name: str) -> types.ColumnName | bool | None:
        """Builds a ColumnName from a raw name.

        In case the name is "true" or "false", returns the equivalent boolean
        (users do not expect to have to capitalize booleans).
        In case the name is "null", returns None.
        """
        if name == "true":
            return True
        if name == "false":
            return False
        if name == types.NULL_REPR:
            return None
        return (
            types.ColumnName(name=self._columns[name], alias=name)
            if name in self._columns
            else types.ColumnName(name=name, alias=name)
        )

    def _parse_expr(self, expr: ast.expr) -> types.Expression:
        """Takes an expression of the ast module and builds an AST in our own format from it."""
        match expr:
            # +LITERAL: +1, +13.37
            case ast.UnaryOp(op=ast.UAdd(), operand=ast.Constant(value=value)):
                return value
            # -LITERAL: -1, -13.37
            case ast.UnaryOp(op=ast.USub(), operand=ast.Constant(value=value)):
                return -value
            # +colname: +mycol, +[my col]
            case ast.UnaryOp(op=ast.UAdd(), operand=ast.Name(id=name)):
                return self._build_name(name)
            # -colname: -mycol, -[my col]
            case ast.UnaryOp(op=ast.USub(), operand=ast.Name(id=name)):
                # Cheating a bit here, assuming the column is numeric
                return types.Operation(left=-1, operator=types.Operator.MUL, right=self._build_name(name))
            # Recursing down into both branches of the operation
            case ast.BinOp(left=left, right=right, op=op):
                operator = self._operator_from_ast_op(op)
                return types.Operation(left=self._parse_expr(left), right=self._parse_expr(right), operator=operator)
            # Constant: number, string literal or boolean
            case ast.Constant(value=value):
                if isinstance(value, bool | int | float | str | None):
                    return value
                else:
                    raise UnsupportedConstant(f"Unsupported constant '{expr}' of type {type(value)}")
            # Column name
            case ast.Name(id=name):
                return self._build_name(name)
            case _:
                raise UnsupportedExpression(f"Unsupported expression: {ast.dump(expr)}")

    def parse(self) -> types.Expression:
        parsed = ast.parse(self.sanitize_formula())
        match parsed:
            case ast.Module(body=[ast.Expr(value=expr)]):
                return self._parse_expr(expr)
            case _:
                raise InvalidFormula("Invalid formula: Expected an expression at root")



from pandas import DataFrame, to_datetime

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import ToDateStep

_FR_TO_EN_MONTHS_LONG = [
    ("janvier", "january"),
    ("février", "february"),
    ("fevrier", "february"),
    ("mars", "march"),
    ("avril", "april"),
    ("mai", "may"),
    ("juin", "june"),
    ("juillet", "july"),
    ("août", "august"),
    ("aout", "august"),
    ("septembre", "september"),
    ("octobre", "october"),
    ("novembre", "november"),
    ("décembre", "december"),
    ("decembre", "december"),
]

_FR_TO_EN_MONTHS_SHORT = [
    ("janv", "jan"),
    ("févr", "feb"),
    ("fevr", "feb"),
    ("mars", "mar"),
    ("avr", "apr"),
    ("mai", "may"),
    ("juin", "jun"),
    ("juil", "jul"),
    ("août", "aug"),
    ("aout", "aug"),
    ("sept", "sep"),
    ("oct", "oct"),
    ("nov", "nov"),
    ("déc", "dec"),
]


def execute_todate(
    step: ToDateStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    format = step.format
    timestamp_unit = None

    if format is None and df[step.column].dtype == "int64":
        # By default int are understood as timestamps, we prefer to convert it to %Y format if values are < 10_000
        if df[step.column].max() < 10_000:
            format = "%Y"
        else:
            # Timestamps are expected in ms (not in ns, which is pandas' default)
            timestamp_unit = "ms"

    if format is not None and ("%b" in format or "%B" in format):
        replacements = _FR_TO_EN_MONTHS_SHORT if "%b" in format else _FR_TO_EN_MONTHS_LONG
        col = df[step.column].str.lower()
        for french_month, english_month in replacements:
            col = col.str.replace(french_month, english_month)
        df[step.column] = col

    datetime_serie = to_datetime(df[step.column], format=format, errors="coerce", unit=timestamp_unit)
    return df.assign(**{step.column: datetime_serie})



from toucan_connectors.common import nosql_apply_parameters_to_query


def nosql_apply_parameters_to_query_with_errors(
    query: dict | list[dict] | tuple | str, parameters: dict | None, **kwargs
):
    """
    When a variable is missing, it may lead to undefined values and therefore pydantic validation errors.
    Instead, we want UndefinedVariableError so we force `handle_errors=True` which does exactly this.
    """
    return nosql_apply_parameters_to_query(query, parameters, handle_errors=True, **kwargs)



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

DATE_GRANULARITY = Literal[
    "day",
    "isoWeek",
    "week",
    "month",
    "quarter",
    "year",
]


class DateGranularityStep(BaseStep):
    name: Literal["dategranularity"] = "dategranularity"
    granularity: DATE_GRANULARITY
    column: str
    new_column: ColumnName | None = None


class DateGranularityStepWithVariable(DateGranularityStep, StepWithVariablesMixin):
    granularity: TemplatedVariable
